"""
Configuration module for access in Google Cloud Platform

Sets JSON file for google environment variable.

Singleton
"""

# pylint: disable=invalid-name, too-few-public-methods

import os
import re
import json
import pathlib
from warnings import warn

from google.auth import default
from google.auth.transport.urllib3 import AuthorizedHttp

def valid_var_env():
    """Checks if var env GOOGLE_APPLICATION_CREDENTIALS is correctly set."""

    return ('GOOGLE_APPLICATION_CREDENTIALS' in os.environ
            and os.environ['GOOGLE_APPLICATION_CREDENTIALS'])

class GCPConf:
    """Singleton configuration class."""

    __instance = None

    class __GCPConf:
        """Internal class."""

        def __init__(self, path, verbose):
            """Construtor da classe interna."""

            self.__valid = False
            self.verbose = verbose
            self.parent = str(pathlib.Path(__file__).parent.parent.absolute())
            self.tempdir = '/tmp'
            self.path = path

            if not self.path and not valid_var_env():
                raise ValueError("GCPConf is singleton and the first instance must "
                                 "pass a valid 'path' to a JSON containing access "
                                 "keys to your service account.\nFallback to env "
                                 "variable GOOGLE_APPLICATION_CREDENTIALS did not "
                                 "work.")
            elif valid_var_env():
                self.path = os.environ['GOOGLE_APPLICATION_CREDENTIALS']
                warn("Param. 'path' invalid. Fallback using env var "
                     "GOOGLE_APPLICATION_CREDENTIALS.")

            if not os.path.exists(self.path):
                raise ValueError("No JSON found in '{}'. "
                                 "Please create a 'GCPConf' object"
                                 " with a valid path".format(self.path))
            else:
                os.environ['GOOGLE_APPLICATION_CREDENTIALS'] = self.path
                self.__valid = True

            self.__print('Instantiating new GCPConf in {}'.format(self.path))

            # set credentials
            self.scopes = ['https://www.googleapis.com/auth/cloud-platform']
            self.credentials, self.project_id = default(scopes=self.scopes)
            self.std_region = 'southamerica-east1'
            # call this only after std_region
            self.std_zone_uri = self.__get_std_zone()
            self.std_zone = self.get_zone_name(self.std_zone_uri)
            # if both are None inform a standard
            if self.std_zone is None:
                self.std_zone = 'southamerica-east1-a'
            if self.std_zone_uri is None:
                self.std_zone_uri = ('https://www.googleapis.com/compute/v1/'
                                     'projects/gm-mateus-estat/zones/southamerica'
                                     '-east1-a')

        def __print(self, msg):
            """Prints message if verbose is set."""

            if self.verbose:
                print(msg)

        def __bool__(self):
            """Returns true if path is valid."""

            return self.__valid

        @staticmethod
        def remove(filename):
            """Removes a file."""

            return os.remove(filename)

        def request(self, method, url, fields=None, headers=None):
            """Make a request using urllib3 AuthorizedHttp."""

            authed_http = AuthorizedHttp(self.credentials)
            return authed_http.request(method, url, fields=fields, headers=headers)

        def get(self, url, fields=None, headers=None):
            """Make a GET request using urllib3 AuthorizedHttp."""

            return self.request('GET', url, fields=fields, headers=headers)

        def __get_std_zone(self):
            """
            This method must be called after self.std_region and self.project_id
            are set.
            """

            url = 'https://compute.googleapis.com/compute/v1/projects/{}/regions/{}'

            try:
                response = self.get(url.format(self.project_id, self.std_region))
                data = json.loads(response.data)

                if response.status == 200:
                    if not data['zones']:
                        self.__print('No zone available for region {}'.format(
                            self.std_region))
                        return None

                    return data['zones'][0]  # returns the first zone
                else:  # something wrong happened
                    warn('Configuration works but is limited: {}'.format(
                        data['error']['message']))
                    return None
            except Exception as excp:
                self.__print('Failed to retrieve region information: {}'.format(
                    excp))
                return None

        @staticmethod
        def get_zone_name(std_zone_uri):
            """Get zone name by zone URI."""

            if std_zone_uri is None:
                return None

            match = re.match('.+/zones/(.+)', std_zone_uri)
            return match.group(1)  # if this fails, the URI has changed

    def __init__(self, path=None, verbose=False):
        """Construtor singleton."""

        if not GCPConf.__instance:
            GCPConf.__instance = GCPConf.__GCPConf(path, verbose)
        elif verbose:
            print('Using same GCPConf in {}'.format(GCPConf.__instance.path))

    def __getattr__(self, attr):
        """Aponta atributos não existentes para os atributos da classe interna."""

        # se a class interna também não tiver, vai lançar uma exceção de atributo
        return getattr(self.__instance, attr)



"""
Represents a Google Cloud Dataproc Workflow
"""

# pylint: disable=too-few-public-methods, too-many-arguments
# pylint: disable=too-many-instance-attributes

import re
import json
from warnings import warn

from google.cloud import dataproc_v1
from google.cloud.dataproc_v1.gapic.transports import \
    workflow_template_service_grpc_transport as wtsgt
from google.cloud.dataproc_v1.proto import workflow_templates_pb2
from google.protobuf import empty_pb2
from google.api_core.operation import from_http_json
from google.api_core.exceptions import GoogleAPICallError
from googleapiclient import discovery

from gcpinfra.gcp.conf import GCPConf
from gcpinfra.gcd.jobs import GCDJob
from gcpinfra.gce.clusters import GCECluster

class GCDWorkflow:
    """
    Google Cloud Dataproc Inline Workflow.

    Creates an ephemeral cluster to run DAGs. After everything is processed, the
    cluster will end its life cycle.

    Please note that jobs are not sequential unless you configure it with a list of
    'prerequisiteStepIds'. Refer to http://bit.ly/google-job-prerequisiteStepIds
    for more details.
    """

    def __init__(self, jobs=None, cluster=None, verbose=False, opname=None,
                 sync=True, callback=None, **kwargs):
        """
        Instantiate an inline workflow.

        This workflow will not be saved with a name. The only way to access it from
        outside is using the operation name.

        This classe requires that either 'jobs' and 'cluster' or 'opname' are
        provided. This means that if you pass 'opname' you cannot pass 'jobs' or
        'cluster'.

        When 'opname' is provided this class represents a previously created
        workflow and will raise an exception if name isn't found in dataproc.
        Param 'opname' must respect the template:
        projects/{project_id}/regions/{region}/operations/{op_uuid}

        Args:
            jobs (list[gcpinfra.gcd.jobs.GCDJob]): Job list to run when
                the cluster starts.
            cluster (gcpinfra.gce.cluster.GCECluster): Cluster that will be created
                and deleted after jobs finish.
            verbose (bool): Optional. Starts in verbose mode.
            opname (str): Optional. Operation name to retrieve the inline workflow
            sync (bool): Optional. Indicating if method 'run' is sync
            callback (callable): Optional. Callable to be invoked when async.
                First arg must always be a bool indicating whether or not the
                execution failed. Second arg must always the result. In case of
                failure, the result will be None. The result type depends on the
                job you executed.
            kwargs (dict): Optional. Params to pass to callback
        """

        self.conf = GCPConf(verbose=verbose)  # must configure this first
        self.verbose = verbose
        self.opname = opname
        self.result = None
        self.sync = sync
        self.callback = callback
        self.callbackkwargs = kwargs
        self.failed = None

        if not self.sync and callback is not None and not callable(callback):
            raise ValueError("Arg 'callback' must be callable.")

        if self.sync and callback is not None:
            warn("Provided arg 'callback' will not be used since arg 'sync'"
                 " is True")

        if opname is not None and any([x is not None for x in [jobs, cluster]]):
            raise ValueError("Arg 'opname' provided with args 'jobs' and 'cluster'."
                             " You can only pass either 'jobs' and 'cluster'"
                             " or 'opname'.")
        elif opname is None:  # use jobs and cluster
            if not isinstance(cluster, GCECluster):
                raise ValueError("Arg 'cluster' must be an instance of "
                                 "'gcpinfra.gcecluster.GCECluster'")

            self.jobs = self.__validate_jobs(jobs)
            self.cluster = cluster

            self.zone_uri = self.cluster.zone_uri
            self.region = self.__get_region()
            self._wf = None
            self._origin_user = True
        else:  # use opname
            self.jobs = None
            self.cluster = None
            self.zone_uri = None
            self.region = self.__get_region_from_opname()
            self.dataproc_workflow_client = None
            self._wf = self.__get_operation_dict()
            self._origin_user = False

        self.dataproc_workflow_client = self.__get_workflow_client(self.region)

    def __print(self, msg):
        """Print msg accordingly to verbose parameter."""

        if self.verbose:
            print(msg)

    def __get_region_from_opname(self):
        """Retrieve the region from the opname."""

        return re.match(r'.*/regions/(.+)/operations.*', self.opname).group(1)

    def __get_operation_dict(self):
        """Returns a dict representing the operation using the opname."""

        self.__print("Retrieving operation '{}'".format(self.opname))
        s = discovery.build('dataproc', 'v1', credentials=self.conf.credentials)
        opdict = s.projects().regions().operations().get(name=self.opname).execute()
        return from_http_json(opdict, self.conf.request, empty_pb2.Empty,
                              metadata_type=workflow_templates_pb2.WorkflowTemplate)

    def __validate_jobs(self, jobs):
        """Verifies if jobs are valid."""

        msg = "Param 'jobs' must be either 'list' of, or single, "
        msg += "'gcpinfra.gcdjobs.GCDJob'"
        if not isinstance(jobs, (GCDJob, list)):
            raise ValueError(msg)

        if not isinstance(jobs, list):  # there is a single job passed
            jobs = [jobs]

        for job in jobs:
            if not isinstance(job, GCDJob):
                raise ValueError(msg)

        return jobs

    def __get_region(self):
        """Returns the region based on the zone."""

        if self.zone_uri is None:
            self.cluster.zone_uri = self.conf.std_zone_uri
            self.zone_uri = self.conf.std_zone_uri

        zone = self.conf.get_zone_name(self.zone_uri)
        try:
            region = '-'.join(zone.split('-')[:-1])
            self.__print("Using region '{}'".format(region))
            return region
        except:
            raise ValueError("Invalid zone '{}'".format(zone))

    def __get_workflow_client(self, region):
        """Based on the region, returns the dataproc workflow client."""

        if region == 'global':  # use the global configuration
            self.__print("Using global region configuration")
            return dataproc_v1.WorkflowTemplateServiceClient()

        client_transport = wtsgt.WorkflowTemplateServiceGrpcTransport(
            address="{}-dataproc.googleapis.com:443".format(region))
        return dataproc_v1.WorkflowTemplateServiceClient(client_transport)

    def mount_json_representation(self, return_type='rep'):
        """Generates JSON representation for the workflow."""

        rep = {
            'placement': {'managed_cluster': self.cluster.get_rep()},
            'jobs': [j.get_rep() for j in self.jobs]
        }

        if return_type == 'rep':
            return rep
        elif return_type == 'str':
            return json.dumps(rep)

        raise ValueError("Param. 'return_type' must be either 'rep' or 'str'.")

    def __async_result(self, future, sync='async'):
        """Get async result from worflow execution."""

        try:
            self.result = future.result()
            self.failed = False
        except GoogleAPICallError:
            self.result = None
            self.failed = True

        if self.callback is not None and not self.sync:
            self.callback(self.failed, self.result, **self.callbackkwargs)

        result = 'success' if self.result is not None else 'failed'
        self.__print('[{}] Workflow finished ({})'.format(sync, result))

    def run(self, sync=None):
        """
        Execute this workflow.
        """

        if sync is not None:
            warn("Arg 'sync' of method 'run' is deprecated. Please use arg"
                 " 'sync' in the constructor.", DeprecationWarning)

        ssync = 'sync' if self.sync else 'async'
        if self._origin_user:
            self.__print('Initiating workflow [{}]. Check '
                         'https://console.cloud.google.com/dataproc/'
                         'workflows/'.format(ssync))

            parent = 'projects/{}/regions/{}'.format(self.conf.project_id,
                                                     self.region)
            wf = self.dataproc_workflow_client.instantiate_inline_workflow_template(
                parent, self.mount_json_representation())
            self._wf = wf
            self.opname = self._wf.operation.name

        if self.sync:
            self.__async_result(self._wf, sync=ssync)
            return self.result
        else:
            self._wf.add_done_callback(self.__async_result)
            return None

    def running(self):
        """Return true in case the workflow is running."""

        if self._wf is not None:
            return self._wf.running()

        return False

    def done(self):
        """Return true in case the workflow is done."""

        if self._wf is not None:
            return self._wf.done()

        return False



"""
GCS Communicator.

Upload and manage files. This module configures checksum check using CRC32c hash
for upload verification

In case of errors, re-raise exceptions sent by Google.
"""

# pylint: disable=invalid-name, try-except-raise

import re
import base64
from hashlib import md5
from datetime import datetime

import pandas as pd
from google.cloud import storage
from crc32c import crc32

from gcpinfra.gcp.conf import GCPConf

class GCSComms:
    """
    GCS Communication Manager
    """

    def __init__(self, bucketname, verbose=False):
        """Constructor."""

        self.verbose = verbose
        self.conf = GCPConf(verbose=verbose)
        self.bucketname = bucketname
        self.storage_client = storage.Client()
        self.bucket = self.storage_client.bucket(bucketname)

        self.decode_to = 'utf-8'

    def __print(self, msg):
        """Prints message accordingly to 'verbose' param."""

        if self.verbose:
            print(msg)

    def __perform_crc32c_on_file(self, filename):
        """Creates CRC32c hash for a file."""

        with open(filename, 'rb') as _f:
            checksum = crc32(_f.read())
            base64_crc32c = base64.b64encode(
                checksum.to_bytes(length=4, byteorder='big')).decode(self.decode_to)
        return base64_crc32c

    def __perform_crc32c_on_str(self, _str):
        """Creates CRC32c hash for a string."""

        checksum = crc32(_str.encode())
        base64_crc32c = base64.b64encode(
            checksum.to_bytes(length=4, byteorder='big')).decode(self.decode_to)
        return base64_crc32c

    def __replace_policy(self, blob, gcsfilename, policy):
        """
        Specify actions.

        Please note the credentials specified in GCPConf must have GCS permissions.

        Policies:
            - 'raise' will delegate the file's override authority to GCS
            - 'ignore' in case the file exists do nothing
            - 'replace' deletes the existing file and uploads a new one
        """

        if policy == 'raise':
            return True

        exists = self.file_exists(gcsfilename=gcsfilename, blob=blob)
        if policy == 'ignore':
            if exists:
                self.__print('File already exists and upload will be ignored')
                return False
            return True

        if policy == 'replace':
            if exists:
                self.delete_file(gcsfilename=gcsfilename, blob=blob)
                self.__print('File already exists and will be replaced')
            return True

        raise ValueError("Invalid policy '{}'".format(policy))

    def get_file_rep(self, gcsfilename):
        """Returns file blob."""

        blob = self.bucket.get_blob(gcsfilename)
        if blob is None:
            return self.bucket.blob(gcsfilename)
        return blob

    def file_exists(self, gcsfilename=None, blob=None):
        """Check if file exists in bucket."""

        if blob is None:
            blob = self.get_file_rep(gcsfilename)

        return blob.exists()

    def delete_file(self, gcsfilename=None, blob=None):
        """
        Delete file from bucket.

        If file does not exist, raises an exception
        """

        if blob is None:
            blob = self.get_file_rep(gcsfilename)

        blob.delete()

    def find_blobs_by_prefix(self, prefix, delimiter=None):
        """
        Find blobs by prefix. One can also pass a delimiter to emulate
        hierarchy.

        In GCS there is no such thing as a directory. Dirs or folders are just
        prefixes of objects

        Also bear in mind that this method gets slower the more objects are
        contained in the prefix.
        """

        _blobs = self.storage_client.list_blobs(self.bucketname, prefix=prefix,
                                                delimiter=delimiter)
        return list(_blobs)

    def find_blobs_by_regex(self, regex, flags=0, prefix=None, delimiter=None):
        """
        Find blobs by refex. Use prefix and delimiter to speed up the process
        and travel less data over the network.

        In GCS there is no such thing as a directory. Dirs or folders are just
        prefixes of objects

        Also bear in mind that this method gets slower the more objects are
        contained in the bucket.
        """

        _all = self.find_blobs_by_prefix(prefix, delimiter=delimiter)
        pttn = re.compile(regex, flags=flags)
        res = []
        for blob in _all:
            if pttn.search(blob.name):
                res.append(blob)
        return res

    def delete_multiples(self, bloblist):
        """
        Delete multiples objects from bucket.

        Also bear in mind that this method gets slower the more objects are
        contained in the list.

        The list must contain just valid blobs.
        """

        for blob in bloblist:
            blob.delete()

    def upload_file(self, filename, gcsfilename, replacepolicy='raise'):
        """Uploads a file."""

        crc32c = self.__perform_crc32c_on_file(filename)
        self.__print('Uploading file {} to {}'.format(filename, gcsfilename))
        blob = self.get_file_rep(gcsfilename)
        if self.__replace_policy(blob, gcsfilename, replacepolicy):
            blob.crc32c = crc32c
            blob.upload_from_filename(filename)
            self.__print('File {} uploaded'.format(gcsfilename))

    def upload_dataframe(self, df, gcsfilename, replacepolicy='raise',
                         extension='parquet'):
        """
        Uploads a pandas dataframe.

        Param. 'df' should be a pandas.DataFrame
        If 'parquet' extension is selected, a temp file will be created and uploaded
        to GCS.
        """

        if not isinstance(df, pd.DataFrame):
            raise ValueError("Param. 'df' should be a 'pandas.DataFrame'.")

        if extension == 'csv':
            self.__print('Uploading {} rows to csv in {}'.format(
                df.shape[0], gcsfilename))
            blob = self.get_file_rep(gcsfilename)
            _str = df.to_csv(index=False)
            crc32c = self.__perform_crc32c_on_str(_str)
            if self.__replace_policy(blob, gcsfilename, replacepolicy):
                blob.crc32c = crc32c
                blob.upload_from_string(_str, content_type='text/plain')
                self.__print('File {} uploaded'.format(gcsfilename))
        elif extension == 'parquet':
            filename = '{}/temp{}.parquet'.format(
                self.conf.tempdir, md5(str(datetime.now()).encode()).hexdigest())
            df.to_parquet(filename)
            try:
                self.upload_file(filename, gcsfilename, replacepolicy)
            except:
                raise
            finally:
                self.conf.remove(filename)
        else:
            raise ValueError("Not supported extension '{}'".format(extension))



"""
Core object for GCP classes.
"""

import json

VALID_MACHINES = [  # beta machines (E2) are not yet ready to be used here
    'n1-standard-1',
    'n1-standard-2',
    'n1-standard-4',
    'n1-standard-8',
    'n1-standard-16',
    'n1-standard-32',
    'n1-standard-64',
    'n1-standard-96',
    'n1-highmem-2',
    'n1-highmem-4',
    'n1-highmem-8',
    'n1-highmem-16',
    'n1-highmem-32',
    'n1-highmem-64',
    'n1-highmem-96',
    'n1-highcpu-2',
    'n1-highcpu-4',
    'n1-highcpu-8',
    'n1-highcpu-16',
    'n1-highcpu-32',
    'n1-highcpu-64',
    'n1-highcpu-96',
    'n2-standard-2',
    'n2-standard-4',
    'n2-standard-8',
    'n2-standard-16',
    'n2-standard-32',
    'n2-standard-48',
    'n2-standard-64',
    'n2-standard-80',
    'n2-highmem-2',
    'n2-highmem-4',
    'n2-highmem-8',
    'n2-highmem-16',
    'n2-highmem-32',
    'n2-highmem-48',
    'n2-highmem-64',
    'n2-highmem-80',
    'n2-highcpu-2',
    'n2-highcpu-4',
    'n2-highcpu-8',
    'n2-highcpu-16',
    'n2-highcpu-32',
    'n2-highcpu-48',
    'n2-highcpu-64',
    'n2-highcpu-80'
]

VALID_JOBS = [
    'hadoop_job',
    'spark_job',
    'pyspark_job',
    'hive_job',
    'pig_job',
    'spark_sql_job'
]

class GCPCore:
    """Base class for Google Compute Engine and Google Cloud Dataproc classes."""

    def __init__(self, objname=None):
        """Constructor."""

        self.objname = objname
        self.__myattrbs = {}

    def add_attr(self, varname, value):
        """Add a attribute."""

        setattr(self, varname, value)
        value = value.get_rep() if isinstance(value, GCPCore) else value
        if isinstance(value, list):
            value = [v.get_rep() if isinstance(v, GCPCore) else v for v in value]
        self.__myattrbs[varname] = value

    def get_rep(self):
        """
        Returns the current representation.

        This method will oftenly be overrided.
        """

        return self.__myattrbs

    def to_json(self):
        """Returns a json representation."""

        if self.objname is None:
            representation = self.__myattrbs
        else:
            representation = {self.objname: self.__myattrbs}

        return json.dumps(representation)

    def __str__(self):
        """to string."""

        return self.to_json()



"""
Represents a job to be run in a cluster or machine.

Here you will find implementations for each type of job.
"""

# pylint: disable=invalid-name, too-few-public-methods, too-many-arguments

from gcpinfra.gcp.core import GCPCore, VALID_JOBS

def valid_job_type(job_type):
    """Checks if the informed job type is valid or not."""

    return job_type in VALID_JOBS

def check_param_type(p, t, ignore_none=False, param_name=''):
    """Checks if the informed param is the correct type."""

    if ignore_none and p is None:
        return

    if not isinstance(p, t):
        raise ValueError("Param. '{}' must be a '{}'".format(param_name, t))

def check_list_type(l, t, ignore_none=False, param_name=''):
    """Checks if only the correct type is present in list values"""

    if ignore_none and l is None:
        return

    if any(map(lambda x: not isinstance(x, t), l)):
        raise ValueError("Param. '{}' must be filled with values of '{}'".format(
            param_name, t))

def nelnn(param):
    """Return true if 'param' is neither an empty list or none."""

    if not isinstance(param, list) and param is not None:
        return True

    return bool(param)

class GCDJob(GCPCore):
    """Generic Job."""

    def __init__(self, step_id, job_type):
        """Constructor."""

        if not valid_job_type(job_type):
            raise ValueError("Param 'job_type' is not valid. Check "
                             "https://cloud.google.com/dataproc/docs/reference/"
                             "rest/v1/projects.locations.workflowTemplates"
                             "#orderedjob for detailed information.")

        super().__init__()
        self.add_attr('step_id', step_id)
        self.job_type = job_type

    def add_conf(self, conf):
        """Adds job configuration."""

        check_param_type(conf, dict, param_name='conf')
        conf = {key: conf[key] for key in conf if nelnn(conf[key])}
        self.add_attr(self.job_type, conf)

class GCDPySparkJob(GCDJob):
    """
    PySpark Job.

    More info: https://cloud.google.com/dataproc/docs/reference/rest/v1/PySparkJob
    """

    def __init__(self, step_id, main, args=None, python_files=None, jar_files=None,
                 other_files=None, archive_files=None, properties=None):
        """Constructor."""

        check_param_type(main, str, param_name='main')  # required
        check_param_type(step_id, str, param_name='step_id')  # required

        # optionals
        check_param_type(args, list, ignore_none=True, param_name='args')
        check_param_type(jar_files, list, ignore_none=True, param_name='jar_files')
        check_param_type(python_files, list, ignore_none=True,
                         param_name='python_files')
        check_param_type(other_files, list, ignore_none=True,
                         param_name='other_files')
        check_param_type(archive_files, list, ignore_none=True,
                         param_name='archive_files')
        check_param_type(properties, dict, ignore_none=True,
                         param_name='properties')

        # check for correct internal type
        check_list_type(args, str, ignore_none=True, param_name='args')
        check_list_type(jar_files, str, ignore_none=True, param_name='jar_files')
        check_list_type(python_files, str, ignore_none=True,
                        param_name='python_files')
        check_list_type(other_files, str, ignore_none=True,
                        param_name='other_files')
        check_list_type(archive_files, str, ignore_none=True,
                        param_name='archive_files')

        conf = {
            'main_python_file_uri': main,
            'args': args,
            'python_file_uris': python_files,
            'jar_file_uris': jar_files,
            'file_uris': other_files,
            'archive_uris': archive_files,
            'properties': properties
        }

        super().__init__(step_id, 'pyspark_job')
        self.add_conf(conf)
